import sys
import pygame
import OpenGL.GL


# Positioning section
# Functions to calculate the position of things within the screen

def fillscreenscale(context, ratio):
    screenratio = context['screenratio']
    if screenratio < ratio:
        return screenratio / ratio
    else:
        return 1.0

def centerx_leftmargin(context, width, ratio, margin):
    return -1.0 + margin + width / 2.0

def centerx_rightmargin(context, width, ratio, margin):
    return  1.0 - margin - width / 2.0

def centery_topmargin(context, width, ratio, margin):
    return  context['screenratio'] - margin - width * ratio / 2.0

def centery_bottommargin(context, width, ratio, margin):
    return -context['screenratio'] + margin + width * ratio / 2.0



# Main section
# main loop pycomb function


def rungame(worldfunc):
    pygame.mixer.pre_init(44100, -16, 2, 1024)
    pygame.init()
    try:
        pygame.display.set_mode((0,0), pygame.FULLSCREEN | pygame.DOUBLEBUF | pygame.OPENGL )
        (width, height) = pygame.display.get_surface().get_size()
    except:
        pygame.display.init()
        (width, height) = pygame.display.list_modes()[0]
        pygame.display.set_mode((width,height), pygame.FULLSCREEN | pygame.DOUBLEBUF | pygame.OPENGL)
    pygame.mouse.set_visible(0)

    context = { 'screenratio': 1.0*height/width, 'state': None }
    OpenGL.GL.glMatrixMode(OpenGL.GL.GL_PROJECTION)
    OpenGL.GL.glLoadIdentity()
    OpenGL.GL.glFrustum(-1.0, 1.0, -context['screenratio'], context['screenratio'], 1.0, 10000.0)

    OpenGL.GL.glEnable( OpenGL.GL.GL_TEXTURE_2D )
    OpenGL.GL.glEnable( OpenGL.GL.GL_BLEND )
    OpenGL.GL.glBlendFunc( OpenGL.GL.GL_SRC_ALPHA, OpenGL.GL.GL_ONE_MINUS_SRC_ALPHA )

    OpenGL.GL.glEnable( OpenGL.GL.GL_VERTEX_ARRAY )
    OpenGL.GL.glEnable( OpenGL.GL.GL_NORMAL_ARRAY )

    while 1:
        OpenGL.GL.glClear( OpenGL.GL.GL_COLOR_BUFFER_BIT | OpenGL.GL.GL_DEPTH_BUFFER_BIT )
        context['time'] = pygame.time.get_ticks()/1000.0
        context['events'] = pygame.event.get()

        context = worldfunc(context)
        if context['state'] == None: sys.exit()

        pygame.display.flip()
