import pygame
import pygame.sndarray
import numpy
import StringIO  # In case of old pygame version,
import struct    # use these two to pack the audio


def generatesound(seconds, soundfunc):
    samples, _, channels = pygame.mixer.get_init()
    t = numpy.linspace(0.0, seconds, samples * seconds)
    v = soundfunc(t)
    if channels > 1:
        vstereo = v.repeat( channels ).reshape(samples*seconds, channels)
    else:
        vstereo = v
    packedsound = numpy.array( 32767.0*vstereo, dtype='h' )
    try:
        return pygame.sndarray.make_sound( packedsound )
    except:  # In case of old pygame version
        bytes = 2
        bits = 16
        data = packedsound.tostring()
        subchunk2 = 'data' + struct.pack('<i', len(data)) + data
        subchunk1 = 'fmt ' + struct.pack('<ihhiihh',
                                           16, 1, channels, samples, bytes*samples*channels, bytes*channels, bits)
        header = 'RIFF' + struct.pack('<i', 4 + (8 + len(subchunk1)) + (8 + len(subchunk2)) ) + 'WAVE'
        return pygame.mixer.Sound( StringIO.StringIO( header + subchunk1 + subchunk2 ) )


def playsound(sound):
    sound.play()


def sinewave(frequency):

    def function(t):
        return numpy.sin(2*numpy.pi*frequency(t)*t)

    return function


def squarewave(frequency):

    def function(t):
        return numpy.where( numpy.modf( t * frequency(t) )[0] > 0.5 , 1.0, -1.0 )

    return function


def multiply_waves(f, g):

    def function(t):
        return ( (f(t) + 1.0) * (g(t) + 1.0) )/2.0 - 1.0

    return function


def frequency_slope(frequency, slope):

    def function(t):
        return frequency + slope*t

    return function


def frequency(freq):
    return frequency_slope(freq, 0.0)


def sumoctaves(octaves, sound):

    def function(t):
        sum = numpy.zeros( t.size )
        factor = 1.0
        for i in range(octaves):
            sum += sound( t * factor ) / (factor * 2.0)
            factor *= 2
        return sum

    return function
