import OpenGL
import pygame

# Aux section
# Auxiliary functions


def joindicts(d,e):
    f = {}
    f.update(d)
    f.update(e)
    return f


def check_keydown(context, key):
    for event in context['events']:
        if event.type == pygame.KEYDOWN:
            if event.key == key:
                return True
    return False


def check_keyup(context, key):
    for event in context['events']:
        if event.type == pygame.KEYUP:
            if event.key == key:
                return True
    return False



# Scene mode section
# functions to set the scene mode


def scene2D(worldfunc):

    def scene(context):
        screenratio = context['screenratio']
        OpenGL.GL.glDisable( OpenGL.GL.GL_DEPTH_TEST ) 
        OpenGL.GL.glMatrixMode( OpenGL.GL.GL_PROJECTION )
        OpenGL.GL.glPushMatrix()
        OpenGL.GL.glLoadIdentity()
        OpenGL.GL.glOrtho(-1.0, 1.0, -screenratio, screenratio, -1.0, 1.0)
        result = worldfunc(context)
        OpenGL.GL.glPopMatrix()
        return result

    return scene


def scene3D(worldfunc):

    def scene(context):
        OpenGL.GL.glEnable( OpenGL.GL.GL_DEPTH_TEST ) 
        OpenGL.GL.glMatrixMode( OpenGL.GL.GL_MODELVIEW )
        OpenGL.GL.glPushMatrix()
        OpenGL.GL.glLoadIdentity()
        result = worldfunc(context)
        OpenGL.GL.glPopMatrix()
        return result

    return scene



# Sequencing section
# Here go the sequencing combinators


def dosteps(steps):

    def sequencer(context):
        stepnumber, substate = context['state'] if context['state'] != None else (0, None)

        result = steps[stepnumber]( joindicts(context, { 'state': substate }) )

        if result['state'] == None:
            stepnumber += 1
            if stepnumber >= len(steps): return result

        return joindicts(result, { 'state': (stepnumber, result['state']) })

    return sequencer


def pipeline(funclist):

    def pipeliner(context):
        statelist = context['state'] if context['state'] != None else [None for x in funclist]

        newstatelist = []
        for worldfunc, state in zip(funclist, statelist):
            context = worldfunc( joindicts(context, {'state': state}) )
            if context['state'] == None: return joindicts(context, {'state': None})
            newstatelist.append(context['state'])

        return joindicts(context, {'state': newstatelist})

    return pipeliner


def switch(condfunc, funclist):

    def switcher(context):
        funcindex, substate = context['state'] if context['state'] != None else (condfunc(context), None)

        result = funclist[funcindex]( joindicts(context, {'state': substate}) )

        if result['state'] == None: return result
        return joindicts(result, { 'state': (funcindex, result['state']) })

    return switcher


def loopwhile(condfunc, worldfunc):

    def looper(context):
        _, substate = context['state'] if context['state'] != None else (True, None)
        if (substate == None) and not condfunc(context): return joindicts(context, {'state': None})

        result = worldfunc( joindicts(context, {'state': substate}) )

        return joindicts(result, { 'state': (True, result['state']) })

    return looper


def loopuntil(condfunc, worldfunc):

    def looper(context):
        _, substate = context['state'] if context['state'] != None else (True, None)

        result = worldfunc( joindicts(context, {'state': substate}) )

        if (result['state'] == None) and condfunc(context): return joindicts(context, {'state': None})
        return joindicts(result, { 'state': (True, result['state']) })

    return looper



# Special combinators
# Combinators that depend on extra parameters / different return values


def generatecontent(progressfunc, contentfuncs, worldfunc):
    contentids = contentfuncs.keys()

    def generator(context):
        position, content, substate = context['state'] if context['state'] != None else (0, {}, None)

        if position <= len(contentids):
            if position < len(contentids):
                key = contentids[position]
                newcontent = joindicts( content, { key: contentfuncs[key](context) } )
            else:
                newcontent = content
            result = progressfunc( joindicts(context,
                                             { 'state': substate, 'extraparams': (position, len(contentids)) }) )
            return joindicts(context, {'state': ((position + 1), newcontent, result['state']) })
        else:
            result = worldfunc( joindicts(context, { 'state': substate, 'resources': content }) )
            if result['state'] == None: return result
            return joindicts(result, {'state': (position, content, result['state']) })

    return generator



# Misc section
# Miscellaneous combinators


def framecount(worldfunc):

    def framecounter(context):
        nframes, starttime, substate = ( context['state'] if context['state'] != None
                                                          else (0,context['time'],None) )

        result = worldfunc( joindicts(context, { 'state': substate } ) )

        if result['state'] == None:
            print nframes / (result['time'] - starttime), "fps"
            return result

        return joindicts(result, { 'state': (nframes + 1, starttime, result['state']) })

    return framecounter


def timecount(message, worldfunc):

    def timecounter(context):
        starttime, substate = context['state'] if context['state'] != None else (context['time'],None)

        result = worldfunc( joindicts(context, { 'state': substate } ) )

        if result['state'] == None:
            print (result['time'] - starttime), message
            return result

        return joindicts(result, { 'state': (starttime, result['state']) })

    return timecounter


def stopkey(key, worldfunc):

    def stopper(context):
        if check_keydown(context, key):
            return joindicts(context, { 'state': None })

        return worldfunc( context )

    return stopper
