import pycomb
import numpy
import OpenGL.GLU
import pygame


# Functions that probably should be added to the pycomb library somewhere


def draw_texture(top, left, right, bottom, texture):
    OpenGL.GL.glEnable( OpenGL.GL.GL_TEXTURE_2D )
    OpenGL.GL.glDisable( OpenGL.GL.GL_COLOR_ARRAY )
    OpenGL.GL.glEnable( OpenGL.GL.GL_TEXTURE_COORD_ARRAY )
    OpenGL.GL.glBindTexture( OpenGL.GL.GL_TEXTURE_2D, texture )
    OpenGL.GL.glVertexPointerd( [ (left, top, 0.0), (right, top, 0.0),
                                  (left, bottom, 0.0), (right, bottom, 0.0) ] )
    OpenGL.GL.glTexCoordPointerd( [ (0.0, 1.0), (1.0, 1.0), (0.0, 0.0), (1.0, 0.0)  ] )
    OpenGL.GL.glDrawElementsui(OpenGL.GL.GL_TRIANGLES, [ (0,1,2), (2,1,3) ])
    OpenGL.GL.glFinish()


def draw_rectangle(top, left, right, bottom, color):
    OpenGL.GL.glDisable( OpenGL.GL.GL_TEXTURE_2D )
    OpenGL.GL.glEnable( OpenGL.GL.GL_COLOR_ARRAY )
    OpenGL.GL.glDisable( OpenGL.GL.GL_TEXTURE_COORD_ARRAY )
    OpenGL.GL.glVertexPointerd( [ (left, top, 0.0), (right, top, 0.0),
                                  (left, bottom, 0.0), (right, bottom, 0.0) ] )
    OpenGL.GL.glColorPointerd( [ color, color, color, color ] )
    OpenGL.GL.glDrawElementsui(OpenGL.GL.GL_TRIANGLES, [ (0,1,2), (2,1,3) ])
    OpenGL.GL.glFinish()


def circle(n, radius, group1, group2, categories, joints):

    def circler(t):
        t2 = numpy.pi * 2.0 * t
        return ( radius * numpy.sin(t2), radius * numpy.cos(t2) )

    return pycomb.mesh.generate2Dmesh(n, circler, group1, group2, categories, joints)



# Functions to generate the game media


def generateExplosion(context):
    mesh = circle(10, 1.0, 'side1', 'side2',
                                        [ {'joints': [('root', 1.0)]} ],
                                        {'root': {'anglex': 0.0, 'angley': 0.0, 'anglez': 0.0, 'length': 1.0 } })
    mesh2 = pycomb.mesh.transform_rotate(mesh, 'side1', (-1.0, 0.0, 0.0), numpy.pi/2.0)
    mesh3 = pycomb.mesh.transform_multiextrude(mesh2, 'side1', [
                                                                   ('side1', 0.5, -0.1, None, 0.0),
                                                                   ('side1', 0.3, -0.3, None, 0.0),
                                                                   ('side1', 0.2, -0.4, None, 0.0)
                                                               ])
    mesh4 = pycomb.mesh.transform_multiextrude(mesh3, 'side2', [
                                                                   ('side2', 0.5, -0.1, None, 0.0),
                                                                   ('side2', 0.3, -0.3, None, 0.0),
                                                                   ('side2', 0.2, -0.4, None, 0.0)
                                                               ])
    mesh5 = pycomb.mesh.remap_trianglegroups(mesh4, {'side1': 'explosion', 'side2': 'explosion'})
    mesh6 = pycomb.mesh.remap_trianglecolors(mesh5, {'explosion': (1.0, 1.0, 0.5, 0.8)})
    return pycomb.mesh.pose(pycomb.mesh.compile(mesh6), {'root':  ( 0.0, 0.0, 0.0) })


def generateMissile(context):
    mesh = circle(10, 0.1, 'base', 'head',
                                        [ {'joints': [('root', 1.0)]} ],
                                        {'root': {'anglex': 0.0, 'angley': 0.0, 'anglez': 0.0, 'length': 1.0 } } )
    mesh2 = pycomb.mesh.transform_multiextrude(mesh, 'head',
                                        [ ('fusselage0', 0.3,  0.0, None, 0.0),
                                          ('fusselage1', 1.0,  0.0, None, 0.0),
                                          ('head', 0.3, -0.05, None, 0.0)]
            )
    mesh3 = pycomb.mesh.regroup_by_normal(mesh2, 'fusselage0', (0.0, 1.0, 0.0), 0.95, 'wing0')
    mesh4 = pycomb.mesh.transform_multiextrude(mesh3, 'wing0', [('wing', 0.1, 0.0, (-1.0, 0.0, 0.0), 0.5)])
    mesh5 = pycomb.mesh.regroup_by_normal(mesh4, 'fusselage0', (0.0, -1.0, 0.0), 0.95, 'wing1')
    mesh6 = pycomb.mesh.transform_multiextrude(mesh5, 'wing1', [('wing', 0.1, 0.0, (1.0, 0.0, 0.0), 0.5)])
    mesh7 = pycomb.mesh.remap_trianglegroups(mesh6, {'wing0': 'wing', 'wing1': 'wing'})
    mesh8 = pycomb.mesh.remap_trianglecolors(mesh7, {
                                                        'base':       (1.0, 1.0, 0.0, 1.0),
                                                        'head':       (1.0, 0.0, 0.0, 1.0),
                                                        'fusselage0': (0.0, 1.0, 0.0, 1.0),
                                                        'fusselage1': (0.0, 0.0, 1.0, 1.0),
                                                        'wing':       (1.0, 0.0, 0.0, 1.0),
                                                    })
    return pycomb.mesh.pose(pycomb.mesh.compile(mesh8), {'root':  ( 0.0, 0.0, 0.0) })


def generateRobot(context):
    mesh = circle(10, 1.0, 'side1', 'side2',
                                        [ {'joints': [('root', 1.0)]} ],
                                        {'root': {'anglex': 0.0, 'angley': 0.0, 'anglez': 0.0, 'length': 1.0 } })
    mesh2 = pycomb.mesh.transform_rotate(mesh, 'side1', (-1.0, 0.0, 0.0), numpy.pi/2.0)
    mesh3 = pycomb.mesh.transform_multiextrude(mesh2, 'side1', [
                                                                   ('body', 1.0, 0.0, None, 0.0),
                                                                   ('body', 0.3, -0.3, None, 0.0),
                                                                   ('head', 0.5, 0.0, None, 0.0),
                                                                   ('head', 0.1, -0.3, None, 0.0),
                                                               ])
    mesh4 = pycomb.mesh.transform_multiextrude(mesh3, 'side2', [
                                                                   ('body', 1.0, 0.0, None, 0.0),
                                                                   ('body', 0.3, -0.3, None, 0.0),
                                                                   ('levitator', 0.5, 0.0, None, 0.0),
                                                                   ('levitator', 0.7, -0.5, None, 0.0),
                                                               ])
    mesh5 = pycomb.mesh.regroup_by_normal(mesh4, 'head', (0.0, 0.0, 1.0), 0.9, 'eye')
    mesh6 = pycomb.mesh.transform_multiextrude(mesh5, 'eye', [('eyewhite', 0.0, -0.2, None, 0.0)])
    mesh7 = pycomb.mesh.remap_trianglegroups(mesh6, {'side1': 'head', 'side2': 'levitator'})
    mesh8 = pycomb.mesh.remap_trianglecolors(mesh7, {
                                                      'head':      (0.3, 0.5, 0.7, 1.0),
                                                      'eye':       (1.0, 0.0, 0.0, 1.0),
                                                      'eyewhite':  (0.3, 0.0, 0.0, 1.0),
                                                      'body':      (0.2, 0.5, 0.5, 1.0),
                                                      'levitator': (0.0, 0.0, 1.0, 1.0),
                                                    })
    return pycomb.mesh.pose(pycomb.mesh.compile(mesh8), {'root':  ( 0.0, 0.0, 0.0) })


def generateMcGee(context):

    def moustache(mesh):
        mesh2 = pycomb.mesh.transform_multiextrude(mesh, 'moustache',
                                            [('moustache', (0.0, 0.2, 0.0), (0.3, 0.0, 0.0), None, 0.0)])
        mesh3 = pycomb.mesh.regroup_by_normal(mesh2, 'moustache', (0.7, 0.0, 0.7), 0.9, 'moustache2')
        mesh4 = pycomb.mesh.transform_displace(mesh3, 'moustache2', (0.0, -0.2, 0.0))
        mesh5 = pycomb.mesh.regroup_by_normal(mesh4, 'moustache', (-0.7, 0.0, 0.7), 0.9, 'moustache3')
        mesh6 = pycomb.mesh.transform_displace(mesh5, 'moustache3', (0.0, -0.2, 0.0))
        return pycomb.mesh.remap_trianglegroups(mesh6, { 'moustache2': 'moustache', 'moustache3': 'moustache' })

    def glasses(mesh):
        return pycomb.mesh.transform_multiextrude(mesh, 'glasses', [ ('glasses', 0.1, 0.1, None, 0.0) ])

    def head(mesh):
        mesh2 = pycomb.mesh.transform_multiextrude(mesh, 'head',
                                                       [ ('neck', 0.2, 0.0, None, 0.0),
                                                         ('neck', 0.1, 0.3, None, 0.0) ]
                )
        mesh3 = pycomb.mesh.transform_scale(mesh2, 'head', (0.7, 1.0, 0.9) )
        mesh4 = pycomb.mesh.transform_displace(mesh3, 'head', (0.0, 0.0, 0.05) )
        mesh5 = pycomb.mesh.transform_multiextrude(mesh4, 'head',
                                                 [('lowerhead', 0.3,  0.1, None, 0.0),
                                                  ('upperhead', 0.3, -0.1, None, 0.0),
                                                  ('head', 0.1, -0.3, None, 0.0)]
                )
        mesh6 = pycomb.mesh.regroup_by_normal(mesh5, 'lowerhead', (0.0, 0.0, 1.0), 0.1, 'moustache')
        mesh7 = pycomb.mesh.regroup_by_normal(mesh6, 'upperhead', (0.0, 0.0, 1.0), 0.1, 'glasses')
        mesh8 = pycomb.mesh.remap_trianglegroups(mesh7,
                  { 'neck': 'head', 'lowerhead': 'head', 'upperhead': 'head' })
        mesh9 = glasses(mesh8)
        mesh10 = moustache(mesh9)
        return mesh10

    def arm(mesh, whicharm, direction):
        armmesh = pycomb.mesh.transform_multiextrude( mesh, whicharm,
                                                 [ ('shoulder', 0.17, 0.0, (0.0, 0.0, direction), numpy.pi/6.0),
                                                   ('shoulder', 0.17, 0.0, (0.0, 0.0, direction), numpy.pi/6.0),
                                                   ('shoulder', 0.17, 0.0, (0.0, 0.0, direction), numpy.pi/6.0),
                                                   ('arm',       1.0, 0.0, None, 0.0),
                                                   ('elbow',    0.17, 0.0, (1.0, 0.0, 0.0), numpy.pi/6.0),
                                                   ('elbow',    0.17, 0.0, (1.0, 0.0, 0.0), numpy.pi/6.0),
                                                   ('elbow',    0.17, 0.0, (1.0, 0.0, 0.0), numpy.pi/6.0),
                                                   ('forearm',   0.3, 0.0, None, 0.0),
                                                   ('gun',       0.0, 0.1, None, 0.0),
                                                   ('gun',       1.0, 0.0, None, 0.0),
                                                   ('gun',       0.0,-0.1, None, 0.0),
                                                   ('gun',      -0.5, 0.0, None, 0.0) ]
                  )
        return pycomb.mesh.remap_trianglegroups(armmesh, { whicharm: 'gun' })

    def torso(mesh):
        mesh2 = pycomb.mesh.transform_multiextrude( mesh, 'torso',
                                                    [ ('lowertorso', 1.0,  0.2, None, 0.0),
                                                      ('lowertorso', 0.5,  0.0, None, 0.0),
                                                      ('uppertorso', 0.5,  0.0, None, 0.0 ),
                                                      ('uppertorso', 0.2, -0.3, None, 0.0),
                                                      ('uppertorso', 0.1, -0.3, None, 0.0) ]
                )
        mesh3 = pycomb.mesh.regroup_by_normal(mesh2, 'uppertorso', ( 1.0, 0.0, 0.0), 0.9, 'leftarm')
        mesh4 = pycomb.mesh.regroup_by_normal(mesh3, 'uppertorso', (-1.0, 0.0, 0.0), 0.9, 'rightarm')
        mesh5 = pycomb.mesh.transform_scale(mesh4, 'torso', numpy.array([0.6, 1.0, 1.6]))
        mesh6 = pycomb.mesh.remap_trianglegroups(mesh5,
                                           {'torso': 'head', 'lowertorso': 'torso', 'uppertorso': 'torso'})
        mesh7 = arm(mesh6, 'leftarm', 1.0)
        mesh8 = arm(mesh7, 'rightarm', -1.0)
        mesh9 = head(mesh8)
        return mesh9

    def saucer(mesh):
        return pycomb.mesh.transform_multiextrude( mesh, 'saucer',
                                                    [ ('saucer', 0.5,  1.5, None, 0.0),
                                                      ('saucer', 0.3,  0.0, None, 0.0),
                                                      ('saucer', 0.5, -1.5, None, 0.0) ]
               )

    def initialmesh():

        mesh = pycomb.mesh.transform_rotate( circle(10, 1.0, 'saucer', 'torso',
                                        [ {'joints': [('root', 1.0)]} ],
                                        {'root': {'anglex': 0.0, 'angley': 0.0, 'anglez': 0.0, 'length': 1.0} } ),
                                     'torso', (1.0, 0.0, 0.0), numpy.pi/2 )
        mesh2 = pycomb.mesh.transform_extrude(mesh, 'torso', 0.0, 'torso_saucer')
        return pycomb.mesh.transform_scale(mesh2, 'torso', numpy.array([0.8, 0.8, 0.5]))

    mesh  = initialmesh()
    mesh2 = torso(mesh)
    mesh3 = saucer(mesh2)

    fleshcolor = (0.99, 0.77, 0.53, 1.0)
    haircolor  = (1.0, 1.0, 1.0, 1.0)
    mesh4 = pycomb.mesh.remap_trianglecolors(mesh3,
                                                { 'torso':    fleshcolor,
                                                  'saucer':   (0.0, 0.0, 1.0, 1.0),
                                                  'torso_saucer':   (0.0, 0.0, 1.0, 1.0),
                                                  'shoulder': fleshcolor,
                                                  'arm':      fleshcolor,
                                                  'elbow':    fleshcolor,
                                                  'forearm':  fleshcolor,
                                                  'gun':      (1.0, 0.0, 0.0, 1.0),
                                                  'head':     fleshcolor,
                                                  'glasses':  (1.0, 1.0, 0.0, 1.0),
                                                  'moustache':haircolor })

    return pycomb.mesh.pose(pycomb.mesh.compile(mesh4), {'root':  ( 0.0, 0.0, 0.0) })


def generateLawn(context):
    return pycomb.texture.texture(512, pycomb.texture.map(
                                           pycomb.texture.tiledperlin([
                                                       [( 1.0,  0.0), ( 0.0,  1.0), ( 0.0,  1.0), (-1.0,  0.0) ],
                                                       [(-1.0,  0.0), ( 0.0, -1.0), ( 0.0,  1.0), ( 0.0,  1.0) ],
                                                       [( 1.0,  0.0), ( 0.0, -1.0), (-1.0,  0.0), ( 0.0, -1.0) ],
                                                       [(-1.0,  0.0), (-1.0,  0.0), ( 0.0,  1.0), ( 0.0, -1.0) ]
                                                                     ]),
                                           pycomb.texture.gradient([
                                                                      (0.0, (0.39, 0.26, 0.13, 1.0)),
                                                                      (0.4, (0.59, 0.46, 0.32, 1.0)),
                                                                      (0.5, (0.0, 0.5, 0.0, 1.0)),
                                                                      (1.0, (0.0, 1.0, 0.0, 1.0))
                                                                   ])
                                                         )
                                 )


def generateShoot(context):
    return pycomb.sound.generatesound(0.2,
               pycomb.sound.sumoctaves(8,
                   pycomb.sound.squarewave( pycomb.sound.frequency_slope(300.0, -100.0) )
               )
           )


def generateBoom(context):
    return pycomb.sound.generatesound(0.5,
               pycomb.sound.sumoctaves(4,
                   pycomb.sound.multiply_waves(
                       pycomb.sound.squarewave( pycomb.sound.frequency_slope(60.0, -6.0) ),
                       pycomb.sound.squarewave( pycomb.sound.frequency_slope(20.0,  2.0) )
                   )
               )
           )


def generateStart(context):
    return pycomb.sound.generatesound(3.0,
               pycomb.sound.sumoctaves(8,
                   pycomb.sound.multiply_waves(
                       pycomb.sound.sinewave( pycomb.sound.frequency_slope(50.0, -10.0) ),
                       pycomb.sound.sinewave( pycomb.sound.frequency_slope(0.0, 10.0) )
                   )
               )
           )



# Functions for the game itself


def loadingbar(message):

    def loader(context):
        position, length = context['extraparams']
        text = (context['state'] if context['state'] != None
                else pycomb.texture.direct_text(message, (0.0, 1.0, 0.0, 1.0), 17, 512))
        pos = float(position) / length
        top, left, right, bottom = 0.02, -0.5, 0.5, -0.02
        delta = 0.02
        draw_rectangle(top + delta, left - delta, right + delta, bottom - delta, (0.0, 0.0, 1.0, 1.0))
        draw_rectangle(top, left, right, bottom, (0.0, 0.5, 1.0, 1.0))
        draw_rectangle(top, left, (1-pos)*left + pos*right, bottom, (0.0, 1.0, 0.0, 1.0))
        draw_texture(bottom, left, right, bottom - (right - left), text)

        if position == length:
            pycomb.texture.closetexture(text)

        return pycomb.state.joindicts(context, { 'state': (None if position == length else text) })

    return loader


def generate_game(size, cellsize, nrobots):
    robot_speed = 1.0
    robot_turning = 0.1

    def random_coordinates(n, size):
        coords = frozenset([])
        while len(coords) < n:
            x,y = int(size * numpy.random.rand()), int(size * numpy.random.rand())
            if (x,y) not in coords:
                coords = coords | frozenset([(x,y)])
        return list(coords)
        

    def newgamestate(context):
        coords = random_coordinates(nrobots + 1, size - 2)
        creatures = ( [('mcgee', ((coords[0][0] + 1) * cellsize, 2.0, (coords[0][1] + 1) * cellsize),
                       numpy.random.rand()*numpy.pi*2, False, 0.0, 0.0)]
                    + [('robot', ((x+1)*cellsize, 3.0, (y+1)*cellsize),
                       numpy.random.rand()*numpy.pi*2, False, robot_turning, robot_speed)
                           for x,y in coords[1:] ] )

        lawn_vertices  = [ (x * cellsize, 0.0, y * cellsize) for x in range(size) for y in range(size) ]
        lawn_normals   = [ (0.0, 1.0, 0.0) for x in range(size) for y in range(size) ]
        lawn_texcoords = [ (float(x), float(y)) for x in range(size) for y in range(size) ]
        lawn_indices   = [ index for x in range(size-1) for y in range(size-1)
                                     for index in (size*x + y,   size*x + y+1,     size*(x+1) + y,
                                                   size*x + y+1, size*(x+1) + y+1, size*(x+1) + y) ]
        lawn = {'texture'      : context['resources']['lawn'],
                'vertexarray'  : numpy.ascontiguousarray(lawn_vertices),
                'normalarray'  : numpy.ascontiguousarray(lawn_normals),
                'texcoordarray': numpy.ascontiguousarray(lawn_texcoords),
                'arrayindices' : numpy.ascontiguousarray(lawn_indices) }

        return {'lawnsize': (size - 1) * cellsize, 'lawn': lawn, 'creatures': creatures }

    def generator(context):
        gamestate = context['gamestate'] if 'gamestate' in context else newgamestate(context)
        return pycomb.state.joindicts(context, { 'state': 1, 'gamestate': gamestate })

    return generator


def display_game():
    explosion_speed = 1.0

    def get_main_character(list, name):
        for creature in list:
            if creature[0] == name:
                return creature
        return None

    def position_camera(radius, declination, ascension, lookat):
        y, z0 = numpy.sin(declination) * radius, numpy.cos(declination) * radius
        x, z = numpy.sin(ascension) * z0, numpy.cos(ascension) * z0
        vy, vz0 = numpy.cos(declination), -numpy.sin(declination)
        vx, vz = numpy.sin(ascension) * vz0, numpy.cos(ascension) * vz0
        cx, cy, cz = lookat
        OpenGL.GLU.gluLookAt(x + cx,y + cy,z + cz, cx, cy, cz, vx,vy,vz)

    def displayer(context):
        OpenGL.GL.glClearColor(0.0, 0.0, 1.0, 1.0)
        OpenGL.GL.glEnable( OpenGL.GL.GL_LIGHTING )
        OpenGL.GL.glEnable( OpenGL.GL.GL_LIGHT0 )
        OpenGL.GL.glEnable( OpenGL.GL.GL_COLOR_MATERIAL )
        OpenGL.GL.glLight(OpenGL.GL.GL_LIGHT0, OpenGL.GL.GL_POSITION, (0.7, -0.7, 0.0, 1.0))

        hero = get_main_character(context['gamestate']['creatures'], 'mcgee')
        if hero == None: return pycomb.state.joindicts(context, {'state': None })
        hero_position, hero_heading = hero[1], hero[2]

        OpenGL.GL.glPushMatrix()
        position_camera(32.0, 0.5, hero_heading + numpy.pi, hero_position)

        pycomb.mesh.render(context['gamestate']['lawn'])

        for creature in context['gamestate']['creatures']:
            modelname                           = creature[0]
            creature_x, creature_y, creature_z  = creature[1]
            creature_heading                    = creature[2]
            creature_exploding                  = creature[3]
            model = modelname if creature_exploding == False else 'explosion'
            OpenGL.GL.glPushMatrix()
            OpenGL.GL.glTranslate( creature_x, creature_y, creature_z)
            OpenGL.GL.glRotate( creature_heading * 180.8 / numpy.pi, 0.0, 1.0, 0.0)
            if creature_exploding != False:
                scale = (context['time'] - creature_exploding) * explosion_speed
                OpenGL.GL.glScale( scale, scale, scale )
            pycomb.mesh.render( context['resources'][ model ] )
            OpenGL.GL.glPopMatrix()

        OpenGL.GL.glPopMatrix()

        return pycomb.state.joindicts(context, {'state': 1 })

    return pycomb.state.scene3D(displayer)


def actuate_game():
    collision_distance = 4.0
    explosion_time     = 2.0
    robot_fire_interval = 1.0
    robot_missile_delta = 4.0
    robot_missile_speed = 20.0


    def think(context, index, creaturelist, distancelist):
        name, (x,y,z), heading, exploding, turnrate, speed = creaturelist[index]
        if (context['time'] % robot_fire_interval) < 0.01:
            pycomb.sound.playsound(context['resources']['shoot'])
            mx = x + numpy.sin(heading) * robot_missile_delta
            mz = z + numpy.cos(heading) * robot_missile_delta
            return [('missile', (mx, y, mz), heading, False, 0.0, robot_missile_speed)]
            
        return []


    def distances(creaturelist):
        list = []
        for i, creature in enumerate(creaturelist):
            list2 = []
            for j, creature2 in enumerate(creaturelist):
                if i == j:
                    list2.append(numpy.inf)
                else:
                    (x1, _, z1) = creature[1]
                    (x2, _, z2) = creature2[1]
                    zd = z2 - z1
                    xd = x2 - x1
                    list2.append( xd*xd + zd*zd)
            list.append(list2)
        return list
                

    def movecreature(context, index, creaturelist, distancelist):
        name, (x,y,z), heading, exploding, turnrate, speed = creaturelist[index]

        if exploding != False:
            if context['time'] - exploding > explosion_time:
                return []
            else:
                return [creaturelist[index]]

        nexploding = False
        for index2, distance in enumerate(distancelist[index]):
            if (distance <= collision_distance) and (creaturelist[index2][3] == False):
                pycomb.sound.playsound(context['resources']['boom'])
                nexploding = context['time']

        previous_time = context['state'] if context['state'] != None else 0.0
        deltat = context['time'] - previous_time
        deltap = speed * deltat
        nx = x + numpy.sin(heading) * deltap
        nz = z + numpy.cos(heading) * deltap
        nheading = heading + turnrate * deltat

        if ((nx < 0.0) or (nz < 0.0)
           or (nx > context['gamestate']['lawnsize']) or (nz > context['gamestate']['lawnsize'])):
            if name == 'missile':
                return []
            else:
                if nz < 0.0:
                    nz = 0.0
                    nheading = numpy.pi - nheading
                if nz > context['gamestate']['lawnsize']:
                    nz = context['gamestate']['lawnsize']
                    nheading = numpy.pi - nheading
                if nx < 0.0:
                    nx = 0.0
                    nheading = 2*numpy.pi - nheading
                if nx > context['gamestate']['lawnsize']:
                    nx = context['gamestate']['lawnsize']
                    nheading = 2*numpy.pi - nheading

        missiles = []
        if name == 'robot':
            missiles = think(context, index, creaturelist, distancelist)

        return [(name, (nx,y,nz), nheading % (numpy.pi*2), nexploding, turnrate, speed)] + missiles

    def actuator(context):
        dist = distances(context['gamestate']['creatures'])
        creatures = [ncreature for index, _ in enumerate(context['gamestate']['creatures'])
                    for ncreature in movecreature(context, index, context['gamestate']['creatures'], dist)]
        gamestate = pycomb.state.joindicts(context['gamestate'], {'creatures': creatures })
        return pycomb.state.joindicts(context, {'state': context['time'], 'gamestate': gamestate })

    return actuator


def interact_game():
    mcgee_acceleration  = 2.0
    mcgee_topspeed      = 10.0
    mcgee_turn_accel    = 0.1
    mcgee_max_turn      = 1.0
    mcgee_missile_speed = 20.0
    mcgee_missile_delta = 4.0
    mcgee_fire_interval = 0.5

    def interact(context, creature):
        name, (x,y,z), heading, exploding, turnrate, speed = creature
        if (name != 'mcgee') or (exploding != False): return [creature]
        previous_time, forward, backward, left, right, fire, nextfire = (context['state']
                      if context['state'] != None else (context['time'], False, False, False, False, False, 0.0))
        deltat = context['time'] - previous_time

        nspeed = speed
        nturn  = turnrate
        if forward and (nspeed < mcgee_topspeed):   nspeed = speed + mcgee_acceleration * deltat
        if backward and (nspeed > -mcgee_topspeed): nspeed = speed - mcgee_acceleration * deltat
        if left and (nturn < mcgee_max_turn):       nturn  = turnrate + mcgee_turn_accel * deltat
        if right and (nturn > -mcgee_max_turn):     nturn  = turnrate - mcgee_turn_accel * deltat

        result = [(name, (x,y,z), heading, exploding, nturn, nspeed)]

        if (not fire) or (context['time'] < nextfire):
            return result
        else:
            pycomb.sound.playsound(context['resources']['shoot'])
            mx = x + numpy.sin(heading) * mcgee_missile_delta
            mz = z + numpy.cos(heading) * mcgee_missile_delta
            missile = [('missile', (mx, y, mz), heading, False, 0.0, mcgee_missile_speed)]
            return result + missile
        

    def interactor(context):
        creatures = [ncreature for creature in context['gamestate']['creatures']
                               for ncreature in interact(context, creature)]
        gamestate = pycomb.state.joindicts(context['gamestate'], {'creatures': creatures })

        _, forward, backward, left, right, fire, nextfire = (context['state'] if context['state'] != None
                                                  else (context['time'], False, False, False, False, False, 0.0))
        if pycomb.state.check_keydown(context, pygame.K_UP):    forward  = True
        if pycomb.state.check_keyup(context, pygame.K_UP):      forward  = False
        if pycomb.state.check_keydown(context, pygame.K_DOWN):  backward = True
        if pycomb.state.check_keyup(context, pygame.K_DOWN):    backward = False
        if pycomb.state.check_keydown(context, pygame.K_LEFT):  left     = True
        if pycomb.state.check_keyup(context, pygame.K_LEFT):    left     = False
        if pycomb.state.check_keydown(context, pygame.K_RIGHT): right    = True
        if pycomb.state.check_keyup(context, pygame.K_RIGHT):   right    = False
        if pycomb.state.check_keydown(context, pygame.K_SPACE): fire     = True
        if pycomb.state.check_keyup(context, pygame.K_SPACE):   fire     = False
        if context['time'] > nextfire: nextfire = context['time'] + mcgee_fire_interval
        return pycomb.state.joindicts(context,
                             {'state': (context['time'], forward, backward, left, right, fire, nextfire),
                              'gamestate': gamestate })

    return interactor


def run_game():
    return pycomb.state.pipeline([
                                     generate_game(20, 10.0, 8),
                                     display_game(),
                                     actuate_game(),
                                     interact_game()
                                 ])


pycomb.game.rungame(
    pycomb.state.stopkey( pygame.K_ESCAPE,
        pycomb.state.generatecontent(
            pycomb.state.scene2D(loadingbar("Please, wait...")),
            { 'mcgee':     generateMcGee,
              'missile':   generateMissile,
              'explosion': generateExplosion,
              'robot':     generateRobot,
              'lawn':      generateLawn,
              'shoot':     generateShoot,
              'boom':      generateBoom,
              'start':     generateStart },
            pycomb.state.framecount(
                pycomb.state.stopkey( pygame.K_q,
                    run_game()
                )
            )
        )
    )
)
